
# --------------------------------------------------------
# - linked list
# - tree
# --------------------------------------------------------
from collections import deque
from itertools import zip_longest, islice


class LinkedListContext:
    @staticmethod
    def transform_args(args, kwargs):
        args = [ListNode.make_linked_list(x) if isinstance(x, list) else x for x in args]
        return args, kwargs


class TreeContext:
    @staticmethod
    def transform_args(args, kwargs):
        args = [TreeNode.make_tree(x) if isinstance(x, list) else x for x in args]
        return args, kwargs


class Context:
    @staticmethod
    def transform_args(args, kwargs):
        return args, kwargs


class ListNode:
    def __init__(self, x=None):
        self.val = x
        self.next = None
        self.has_cycle = False

    def __iter__(self):
        p = self
        while p is not None:
            yield p.val
            p = p.next

    def __str__(self):
        if self.has_cycle:
            return "->".join([str(v) for v in islice(self, 5)]) + '...'
        else:
            return "->".join([str(v) for v in self])

    def __eq__(self, other):
        if self.has_cycle:
            raise TypeError('__eq__ is not supported for linked list who has cycle')
        if not isinstance(other, self.__class__):
            return False
        return all(x == y for x, y in zip_longest(self, other, fillvalue=None))

    @staticmethod
    def make_linked_list(data):
        """make a linked list from a list

        Examples:
        >>> ll = ListNode.make_linked_list([1, 2, 3, 4, 5])
        >>> print(ll)
        1->2->3->4->5

        Args:
            data: data source, an iterable object.

        Returns:
            the head of the linked list, class `LinkedLinkedListNode`.
        """
        data = list(data)
        if len(data) == 0:
            return None
        head = tail = ListNode(data[0])
        for item in data[1:]:
            node = ListNode(item)
            tail.next = node
            tail = node
        return head

    @staticmethod
    def make_cycle_list(data, n):
        """make a linked list (has a cycle) from a list

        Examples:
        >>> cl = ListNode.make_cycle_list([1, 2, 3], 0)
        >>> cl.next.val
        2
        >>> cl.next.next.next.val
        1

        Args:
            data: data source, an iterable object.
            n: link the tail to the nth node. if n < 0 or n > len(data),
               the returned linked list has no cycle.

        Returns:
            the head of the linked list, class `LinkedLinkedListNode`.
        """
        head = ListNode.make_linked_list(data)
        if head is None:
            return None

        if n < 0:
            return head

        # find the tail
        tail = head
        while tail.next:
            tail.has_cycle = True
            tail = tail.next
        tail.has_cycle = True
        # find nth node
        target, cnt = head, 0
        while target:
            if cnt == n:
                break
            cnt += 1
            target = target.next

        # link them
        tail.next = target
        return head


class TreeNode:
    def __init__(self, x=None):
        self.val = x
        self.left = None
        self.right = None

    def __str__(self):
        return "Tree({})".format("-".join([str(v) for v in self]))

    def __iter__(self):
        remaining_nodes = [self]
        fillup = 0
        for node in remaining_nodes:
            if node is None:
                fillup += 1
            else:
                while fillup > 0:
                    yield None
                    fillup -= 1
                yield node.val
                remaining_nodes.append(node.left)
                remaining_nodes.append(node.right)

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return False
        return all(x == y for x, y in zip_longest(self, other, fillvalue=None))

    @staticmethod
    def make_tree(data):
        """make a binary tree from a list

        Examples:
        >>> t = TreeNode.make_tree([1, 2, 3, 4])
        >>> print(t.left)
        Tree(2-4)
        >>> print(t.right)
        Tree(3)

        Args:
            data: a tree node value list in level traversal order

        Returns:
            the root of the binary tree, class `TreeNode`.

        Raises:
            ValueError: If the given data can't be made a vliad binary tree.
        """
        if len(data) < 1:
            return None
        root = TreeNode(data[0])
        queue = deque([root])
        i = 1
        while i < len(data):
            try:
                upper_node = queue.popleft()
            except IndexError:
                raise ValueError("bad data for binary tree")
            for val, child in zip(data[i:i+2], ('left', 'right')):
                if val is not None:
                    new_node = TreeNode(int(val))
                    setattr(upper_node, child, new_node)
                    queue.append(new_node)
            i += 2
        return root



import re
import sys
import json
import time
import logging
import tempfile
from pathlib import Path
from textwrap import indent, shorten
from datetime import datetime
from types import SimpleNamespace

import requests

from leezy.render import Render
from leezy.errors import *
from leezy.utils import SecretDialog, YesNoDialog, SessionTokenDialog
from leezy.config import config, session_token, Urls


LOG = logging.getLogger(__name__)
Info = LOG.info
Debug = LOG.debug
Warn = LOG.warning

ID_WIDTH = 3
NAME_BLACKLIST_RE = re.compile(r'[\\/:.?<>|]')

# this is awesome
GUIDE = """
x  x  | Elements | ... | Memory | > Application < | Security | ...
-----------------------------------------------------------------------------
Application                  | Name | Value | Domain | Path | Expires | ... |
  ...                        |  ..
Storage                      |  ..
  ...                        | LEETCODE_SESSION | ⭐ | .leetcode.com | / | ⭐
  IndexedDB                  |  ..
  Web SQL                    |  ..
  Cookies                    |  ..
   > https://leetcode.com <  |  ..
Cache                        |  ..
   ...                       |  ..
-----------------------------------------------------------------------------
"""


class Payload:
    def __init__(self):
        self.operation = ""
        self.variables = {}
        self.query = ""

    def set_variable(self, key, val):
        self.variables[key] = val

    def as_dict(self):
        return {
            'operationName': self.operation,
            'variables': self.variables,
            'query': self.query
        }


class LoginPayload(Payload):
    def __init__(self):
        self.operation = 'signInWithPassword'
        self.variables = {
            'data': {
                'username': 'example',
                'password': 'no-password'
            }
        }
        self.query = """\
        mutation signInWithPassword($data: AuthSignInWithPasswordInput!) {
            authSignInWithPassword(data: $data) {
                ok
                __typename
            }
        }"""

    def set_secret(self, username, password):
        self.variables['data']['username'] = username
        self.variables['data']['password'] = password
        return self


class ProblemQueryPayload(Payload):
    def __init__(self):
        self.operation = 'questionData'
        self.variables = {
            'titleSlug': 'some-problem'
        }
        self.query = """\
        query questionData($titleSlug: String!) {
            question(titleSlug: $titleSlug) {
                questionId
                questionFrontendId
                title
                titleSlug
                content
                translatedContent
                isPaidOnly
                difficulty
                likes
                dislikes
                similarQuestions
                topicTags {
                    name
                    slug
                }
                companyTagStats
                codeSnippets {
                    langSlug
                    code
                }
                stats
                hints
                status
                sampleTestCase
            }
        }"""

    def set_title_slug(self, title_slug):
        self.variables['titleSlug'] = title_slug
        return self


# just use this to fetch csrftoken
class UserStatusPayload(Payload):
    def __init__(self):
        self.operation = 'userStatus'
        self.variables = {}
        self.query = """\
        query userStatus {
            userStatus {
                __typename
            }
        }"""


class Net:
    def __init__(self):
        self.sess = requests.Session()
        self.sess.headers.update({
            'origin': Urls.portal(),
            'user-agent': ('Mozilla/5.0 (Windows NT 10.0; Win64; x64) '
                           'AppleWebKit/537.36 (KHTML, like Gecko) '
                           'Chrome/80.0.3987.132 Safari/537.36')
        })
        # init csrf once using last local storage
        # `session` will hanlde the upating of csrf caused by multiple requests
        if not session_token.get_csrf():
            # update csrftoken
            Debug('Try initialize csrf')
            r = self.sess.post(Urls.graphql(),
                               json=UserStatusPayload().as_dict())
            found = session_token.try_update_csrf(r)
            if not found:
                Warn('Failed to intialize csrf')
        self.sess.cookies.update(session_token.get_csrf())

    def get(self, url, purpose='', **kwargs):
        self.ensure_login()
        purpose = purpose or f'try to GET {url!r}'
        return self._get(url, purpose=purpose, **kwargs)

    def _get(self, url, purpose='', **kwargs):
        description = purpose
        try:
            r = self.sess.get(url, **kwargs)
        except requests.ConnectionError as err:
            raise NetworkError(description, err)
        # try to record csrf if there is any
        session_token.try_update_csrf(r)
        raise_for_status(r, description)
        return r

    def post(self, url, purpose='', **kwargs):
        self.ensure_login()
        purpose = purpose or f'try to POST {url!r}'

        # add 'x-csrftoken' into POST headers
        # csrf = None
        # for d in ('leetcode.com', '.leetcode-cn.com', ''):
        #     csrf = self.sess.cookies.get('csrftoken', default=None, domain=d)
        #     if csrf:
        #         break
        csrf = self.sess.cookies.get('csrftoken', default=None)
        assert csrf is not None
        headers = {'x-csrftoken': csrf}
        if 'headers' in kwargs:
            kwargs['headers'].update(headers)
        else:
            kwargs['headers'] = headers

        return self._post(url, purpose=purpose, **kwargs)

    def _post(self, url, purpose='', **kwargs):
        description = purpose
        try:
            r = self.sess.post(url, **kwargs)
        except requests.ConnectionError as err:
            raise NetworkError(description, err)
        session_token.try_update_csrf(r)
        raise_for_status(r, description)
        return r

    def ensure_login(self):
        if not session_token.is_existed() or session_token.is_expired():
            if not session_token.is_existed():
                Debug('Session token is not found.')
            else:
                Debug('Session has expired.')
            # login() will update session_token
            Login(self, config.get('core.zone')).login()
        self.sess.cookies.update(session_token.get_token())


class Login:
    def __init__(self, net, zone):
        self.net = net
        if zone == 'cn':
            self.login = self.cn_login
        elif zone == 'us':
            self.login = self.us_login
        else:
            raise ConfigError(f'Unrecognized zone {zone!r}')

    def cn_login(self):
        username, password = self._collect_secret()
        token, expires = self._cn_login_by_secret(username, password)
        session_token.store_token(token, expires)

    def us_login(self):
        Warn('Leetcode.com enables reCAPTCHA verification now.\n'
             'We can\'t sign you in easily anymore. \n'
             'You have to sign in leetcode.com first and then '
             'find LEETCODE_SESSION in the debug panel of browser\n'
             'If you use Chrome: \n'
             '  1. sign in leetcode.com\n'
             '  2. press F12\n'
             '  3. follow the guide of the graph to copy the token out\n'
             + GUIDE)
        token, expires = self._login_by_bare_hand()
        session_token.store_token(token, expires)

    def _login_by_bare_hand(self):
        return SessionTokenDialog().collect()

    def _collect_secret(self):
        try:
            username = config.get('user.name')
            password = config.get('user.password')
        except ConfigError:
            dialog = SecretDialog(f"Sign to {Urls.portal()}")
            username, password = dialog.collect()
        return (username, password)

    def _cn_login_by_secret(self, username, password):
        Debug(f"Try to sign in {Urls.portal()} as {username!r}")
        # leetcode-cn.com
        payload = (LoginPayload().set_secret(username, password).as_dict())
        r = self.net._post(Urls.graphql(),
                           purpose="try to sign in LeetCode",
                           json=payload)

        if not r.json()['data']['authSignInWithPassword']['ok']:
            raise LoginError("Wrong username or password?")
        token = expires = None
        for c in r.cookies:
            if c.name == 'LEETCODE_SESSION':
                token = c.value
                expires = c.expires
                break
        else:
            raise LoginError("login successfully, "
                             "but didn't find session token in the response")
        return (token, expires)


class Entry:
    def __init__(self, question_id, title, title_slug, difficulty):
        self.question_id = question_id
        self.title = title
        self.title_slug = title_slug
        self.difficulty = difficulty


class ProblemEntryRepo:
    def __init__(self):
        self.net = Net()
        path = str(Path(tempfile.gettempdir()) / "leezy_problems.json")
        self.problems_file = path
        self.problems = self._load_local_cache()

    def entry_by_id(self, id_):
        r = self.problems.get(str(id_), None)
        if r is None:
            self._update_cache()
            r = self.problems.get(str(id_), None)
            if r is None:
                raise NotFound(f'Problem<{id_}> is not found')
        return Entry(**r)

    def _load_local_cache(self):
        try:
            f = open(self.problems_file, encoding='utf8')
        except FileNotFoundError:
            return {}
        problems = json.load(f)
        f.close()
        return problems

    def _update_cache(self):
        raw_problems = self._raw_web_all_problems()
        self.problems = self._flush_raw_all_problems(raw_problems)
        self.write_down_problems(self.problems)

    def _raw_web_all_problems(self):
        purpose = "fetch the list of problem entry"
        r = self.net.get(Urls.api_problems_all(), purpose=purpose)
        return r.json()

    def _flush_raw_all_problems(self, raw_json):
        problems = raw_json['stat_status_pairs']
        levels = ['void', 'easy', 'medium', 'hard']
        maps = {}
        for p in problems:
            maps[p['stat']['frontend_question_id']] = {
                "question_id": p['stat']['question_id'],
                "title": p['stat']['question__title'],
                "title_slug": p['stat']['question__title_slug'],
                "difficulty": levels[p['difficulty']['level']]
            }
        return maps

    def write_down_problems(self, problems):
        with open(self.problems_file, 'w', encoding='utf8') as f:
            json.dump(problems, f, ensure_ascii=False)


class ProblemProvider:
    def __init__(self):
        self.entry_repo = ProblemEntryRepo()
        self.net = self.entry_repo.net

    def info_by_id(self, id_):
        """return a `Entry`, provide some basic infomation like title and id
        """
        return self.entry_repo.entry_by_id(id_)

    def detail_by_id(self, id_):
        """return a `dict`, provide code_snippets / description etc, including
        the basic information returned by `info_by_id`
        """
        entry = self.entry_repo.entry_by_id(id_)
        raw = self._raw_problem_detail(entry.title_slug)
        detail = self._flush_raw_problem_detail(entry.title_slug, raw)
        # append basic infomation about this problem
        detail.update(entry.__dict__)
        return detail

    def _raw_problem_detail(self, title_slug):
        payload = ProblemQueryPayload().set_title_slug(title_slug).as_dict()
        purpose = f"fetch problem {title_slug!r} detail"
        post = self.entry_repo.net.post
        r = post(Urls.graphql(), purpose=purpose, json=payload)
        return r.json()

    def _flush_raw_problem_detail(self, title_slug, raw_json):
        if 'errors' in raw_json:
            raise FetchError(
                description=f"found errors when fetching {title_slug!r}",
                cause=raw_json['errors'][0]['message'])
        raw = raw_json['data']['question']
        if raw is None:
            raise FetchError("No content in response.")
        if raw['isPaidOnly']:
            raise Locked(f'the problem {title_slug!r} is locked')
        new = {}
        if 'Please switch to Chinese' in raw['content']:
            new['content'] = raw['translatedContent']
        else:
            new['content'] = raw['content']
        new['similar_problems'] = json.loads(raw['similarQuestions'])
        new['code_snippet'] = [sp['code'] for sp in raw['codeSnippets'] if
                               sp['langSlug'] == 'python'][0].replace('\r\n', '\n')
        # testcase of problem 191 is not valid json data
        # actually, the testcase of 191 is confusing
        # is it ok to abandon it?
        cases = []
        for case_row in raw['sampleTestCase'].split('\n'):
            try:
                cases.append(json.loads(case_row))
            except json.JSONDecodeError:
                pass
        new['sample_testcase'] = cases
        return new


def _find_func_names(text):
    return re.findall(r'def\s+(\S+)\s*\(', text)


class Problem:
    def __init__(self, id_, context=None, provider=None):
        # id_ is frontend_id of a problem
        self.query_id = id_
        self.context = context
        self.provider = provider or ProblemProvider()

        _info = self.provider.info_by_id(id_)
        self.basic_info = _info
        # eliminate all invalid characters before making new files or folders
        # loc means local
        self.loc_id = str(id_).rjust(ID_WIDTH, '0')
        self.loc_title = NAME_BLACKLIST_RE.sub('', _info.title).strip()
        self.loc_slug = NAME_BLACKLIST_RE.sub('', _info.title_slug).strip()
        workdir = Path(config.get('core.workdir'))
        self.folder_path = workdir / Path(f'{self.loc_id} - {self.loc_title}')
        self.html_path = self.folder_path / f'{self.loc_id}.html'
        self.py_path = self.folder_path / f'{self.loc_id}_{self.loc_slug}.py'

        self.content = None
        self.similar_problem = None
        self.code_snippet = None
        self.sample_testcase = None

    def _lazy_init(self):
        detail = self.provider.detail_by_id(self.query_id)
        self.__dict__.update(detail)

    def __str__(self):
        return f'Problem<{self.loc_id}: {self.loc_title}>'

    def _generate_solution_tmpl(self):
        render = Render(self)
        return render.render()

    def digest(self):
        return (f'Problem<{self.loc_id}: {self.loc_title}> '
                f'@{self.basic_info.difficulty}')

    def pull(self):
        if self.content is None:
            self._lazy_init()
        self.folder_path.mkdir(parents=True, exist_ok=True)
        self.html_path.write_text(self.content, encoding='utf8')
        self.py_path.write_text(
            self._generate_solution_tmpl(), encoding='utf8')

    def submit(self, n):
        if not self.py_path.is_file():
            raise LeezyError(f'File not found: {self.py_path}')
        extractor = SolutionExtractor(self.py_path.read_text(encoding='utf8'))
        func, code = extractor.submission(n)
        # if there are multiple solutions,
        # we need to change function name before submitting
        if len(extractor) > 1:
            if self.content is None:
                self._lazy_init()
            origin_func_name = _find_func_names(self.code_snippet)[0]
            code = code.replace(func+'(', origin_func_name+'(')

        prelude = f"Is it OK to submit solution {func!r}?:\n{code}\n"
        if not YesNoDialog(prelude).collect():
            return

        payload = {
            "question_id": str(self.basic_info.question_id),
            "lang": "python3",
            "typed_code": code,
            "test_mode": False,
            "test_judger": "",
            "questionSlug": self.basic_info.title_slug
        }

        net = self.provider.net
        headers = {"referer": Urls.problem_home(self.basic_info.title_slug)}
        r = net.post(Urls.problem_submit(self.basic_info.title_slug),
                     purpose="submit a solution",
                     json=payload,
                     headers=headers)
        submission_id = r.json()['submission_id']
        check_url = Urls.submission_check(submission_id)

        check_cnt = 0
        r = None
        while True:
            time.sleep(1)
            r = net.get(check_url, purpose=f'check submission x {check_cnt}')
            if len(r.json()) > 1:
                break

        rjson = r.json()
        if 'status_code' in rjson:
            # append more infomation
            rjson.update({
                'discuss_url': Urls.problem_discussion(self.basic_info.title_slug),
                'submission_detail': Urls.submission_detail(submission_id),
            })
            SubmissionReporter(rjson).report()
        else:
            raise LeezyError(f'Bad submission: {r.text}')


class Reporter:
    def __init__(self, data):
        self.data = data

    def summary(self):
        print(f'----------------{self.data.status_msg}!----------------')

    def explain(self):
        data = self.data
        print(f'passed cases: {data.total_correct}/{data.total_testcases}')

    def report(self):
        self.summary()
        self.explain()


class RuntimeErrorReporter(Reporter):
    def explain(self):
        print('\nerror:\n', self.data.runtime_error)
        print('\ndetail:\n', self.data.full_runtime_error)


class CompileErrorReporter(Reporter):
    def explain(self):
        print('\nerror:\n', self.data.complie_error)
        print('\ndetail:\n', self.data.full_compile_error)


class WrongAnswerReporter(Reporter):
    def explain(self):
        data = self.data
        print('   passed cases: ',
              f'{data.total_correct}/{data.total_testcases}')
        print('  last testcase: ', shorten(data.input_formatted, 50))
        print('    your output: ', shorten(data.code_output, 50))
        print('expected output: ', shorten(data.expected_output, 50))
        if len(data.input_formatted) < 50 and len(data.expected_output) < 50:
            print('\nconsider to add the following line to your local test?')
            print(f"q.add_case(q.case({data.input_formatted})"
                  f".assert_equal({data.expected_output}))")


class TLEReporter(Reporter):
    def explain(self):
        data = self.data
        print('   passed cases: ',
              f'{data.total_correct}/{data.total_testcases}')


class AcceptedReporter(Reporter):
    def explain(self):
        data = self.data
        print('  time used & rank:',
              f'{data.status_runtime} faster than {data.runtime_percentile:.2f}%')
        print('memory used & rank:',
              f'{data.status_memory} less than {data.memory_percentile:.2f}%')

        print('\nmore helpful links:')
        links = '\n'.join([data.submission_detail, data.discuss_url])
        print(indent(links, '    '))


class SubmissionReporter:
    def __init__(self, data):
        data = SimpleNamespace(**data)
        stat_code = data.status_code
        reporter_map = {
            10: AcceptedReporter,
            11: WrongAnswerReporter,
            14: TLEReporter,
            15: RuntimeErrorReporter,
            20: CompileErrorReporter
        }
        if stat_code not in reporter_map:
            raise LeezyError(f"unkown report type: \n {data!r}")

        r = reporter_map[stat_code](data)
        self.reporter = r

    def report(self):
        self.reporter.report()


class Liner:
    def __init__(self, content):
        self.lines = content.split('\n')
        self.curidx = 0
        self.N = len(self.lines)

    def eat(self):
        self.curidx += 1

    def has_next(self):
        return self.curidx < self.N

    def is_empty(self):
        return self.peek().strip() == ''

    def peek(self):
        return self.lines[self.curidx]

    def indent(self):
        i = 0
        line = self.peek()
        while i < len(line) and line[i] == ' ':
            i += 1
        return i

    def peek_next(self, n):
        nxt = self.curidx + n
        if nxt >= self.N:
            return ''
        else:
            return self.lines[nxt]


class SolutionExtractor:
    def __init__(self, content):
        liner = Liner(content)
        imports = []
        solutions = []
        helpers = []
        while liner.has_next():
            cur_line = liner.peek()
            s_cur = cur_line.strip()
            if s_cur.startswith('import ') or s_cur.startswith('from '):
                if 'leezy' not in cur_line:
                    imports.append(s_cur)
                liner.eat()
            elif '@solution' in liner.peek():
                liner.eat()
                solutions.append(self.get_function(liner))
            elif 'def ' in liner.peek():
                func, code = self.get_function(liner)
                if func != 'main':
                    helpers.append((func, code))
            else:
                liner.eat()

        submits = []
        for solution, code in solutions:
            blocks = [code]
            for helper, helper_code in helpers:
                if helper in code:
                    blocks.append(helper_code)
            parts = imports[:]
            parts.append("\nclass Solution:")
            parts.extend(indent(block, '    ') + '\n' for block in blocks)
            submits.append((solution, '\n'.join(parts)))

        self.submits = submits

    def submission(self, n):
        if n <= 0:
            raise LeezyError('Solution index starts with 1')
        N = len(self.submits)
        be = 'are' if N > 1 else 'is'
        plural = 's' if N > 1 else ''
        if n > N:
            raise LeezyError(f'There {be} only {N} solution{plural}')
        return self.submits[n-1]

    def __len__(self):
        return len(self.submits)

    def __iter__(self):
        yield from self.submits

    def get_function(self, liner):
        indent = liner.indent()
        claim = liner.peek()[indent:]
        func_name = _find_func_names(claim)[0]
        codes = [claim]
        liner.eat()
        while liner.has_next():
            if liner.is_empty():
                codes.append('\n')
            elif liner.indent() > indent:
                codes.append(liner.peek()[indent:])
            else:
                break  # non-empty and smaller indent
            liner.eat()
        # remove trailing new lines
        while codes and codes[-1] == '\n':
            codes.pop()
        return (func_name, '\n'.join(codes) + '\n')



import re
from itertools import zip_longest
from textwrap import wrap, shorten
from getpass import getuser, getpass
from datetime import datetime
from leezy.errors import ConfigError


class SecretDialog:
    def __init__(self, prelude):
        self.prelude = prelude

    def collect(self):
        print(self.prelude)
        username = input("> username: ")
        password = getpass("> password: ")
        return username, password


class YesNoDialog:
    def __init__(self, prelude):
        self.prelude = prelude

    def collect(self):
        print(self.prelude)
        while True:
            answer = input("> [Yes/No]?")
            if answer.upper() in ['Y', 'YES', 'YEAH']:
                return True
            elif answer.upper() in ['N', 'No']:
                return False
            else:
                print(f'< what does {answer!r} mean?')


class SessionTokenDialog:
    def collect(self):
        token = input('> LEETCODE_SESSION token: ')
        print('\n you can set expires using following format:\n',
              '    1.absolute timestamp\n',
              '    2.absolute date like 2020-03-31T00:39:14.945Z\n',
              "    3.relative date like '10s', '11h', '12d'")
        expires = input("> expires: ")

        match = re.findall(r'^(\d+)([dhms]?)$', expires)
        if match:
            num, ty = match[0]
            now_ts = datetime.timestamp(datetime.now())
            num = int(num)
            if ty == '':
                if num <= now_ts:
                    raise ConfigError('Expires is too old')
                expires = num
            elif ty == 'd':
                expires = now_ts + num * 3600 * 24
            elif ty == 'h':
                expires = now_ts + num * 60 * 60
            elif ty == 'm':
                expires = now_ts + num * 60
            elif ty == 's':
                expires = now_ts + num
        else:
            try:
                expires = datetime.strptime(expires, '%Y-%m-%dT%H:%M:%S.%fZ')\
                                  .timestamp()
            except ValueError:
                raise ConfigError('Unrecognized expires formt')
        return (token, int(expires))


class Table:
    """ Table format tool

    Example:
    >>> table = Table()
    >>> table.add_header(['', 'Earth', 'Moon'])
    >>> table.add_row(['Size', '42', '42'])
    >>> print(table)
    +--------+---------+--------+
    |        |  Earth  |  Moon  |
    +========+=========+========+
    |  Size  |  42     |  42    |
    +--------+---------+--------+

    Args(Optional):
        max_col_width: int, max column width
        max_content_width: int, exceeded content will be shortened
    """

    def __init__(self, **kwargs):
        self.__rows = [None]
        self.col_n = 0
        self.max_col_width = int(kwargs.get('max_col_width', 30))
        self.max_content_length = int(kwargs.get('max_content_length', 100))
        self.pad = 2
        self.pad_str = ' ' * 2
        self.col_widths = []
        self.fillup = '<unkown>'
        self.lines = []

    def add_header(self, header):
        self.__rows[0] = [str(h) for h in header]

    def add_row(self, row):
        self.__rows.append([str(x) for x in row])

    def __determine_col_widths(self):
        for colx in zip_longest(*self.__rows, fillvalue=self.fillup):
            self.col_widths.append(
                min(self.max_col_width, max(map(len, colx))))
        self.col_n = max(map(len, self.__rows))
        self.dummy_iter = tuple(range(self.col_n))

    def _format_row(self, row):
        cells = []
        height = 0
        for text, i in zip_longest(row, self.dummy_iter, fillvalue=self.fillup):
            if self.max_content_length > 6:
                text = shorten(text, self.max_content_length)
            col_width = max(1, self.col_widths[i])
            text = wrap(text, col_width)
            height = max(height, len(text))
            cells.append([f"{txt:<{col_width}}" for txt in text])

        for i in range(height):
            line = []
            for j, cell in enumerate(cells):
                if len(cell) <= i:
                    line.append(' ' * self.col_widths[j])
                else:
                    line.append(cell[i])
            pad = self.pad_str
            content = f'{pad}|{pad}'.join(line)
            self.lines.append(f'|  {content}  |')

    def format_sep_line(self, ch='-', sep='+'):
        parts = []
        for w in self.col_widths:
            parts.append(ch * (w + 2 * self.pad))
        self.lines.append(f"{sep}{'+'.join(parts)}{sep}")

    def format_head(self):
        self.format_sep_line()
        self._format_row(self.__rows[0])
        self.format_sep_line(ch='=')

    def format_body(self):
        for row in self.__rows[1:]:
            self._format_row(row)
            self.format_sep_line()

    def __str__(self):
        self.__determine_col_widths()
        self.format_head()
        self.format_body()
        return '\n'.join(self.lines)



import os
import inspect

from pathlib import Path
from collections import defaultdict
from time import perf_counter
from copy import deepcopy
from enum import Enum
from io import StringIO
from textwrap import shorten, dedent

import pytest

from leezy.utils import Table
from leezy.config import config
from leezy.assists import Context


def solution(func):
    """Attach the `func` a solution marker
    """
    func.__dict__['solution'] = True
    return func


def timeit(func):
    """Attach the `func` a time marker
    """
    func.__dict__['timeit'] = True
    func.__dict__['precision'] = 4
    return func


def timeit_with_precision(precision):
    def inject(func):
        func.__dict__['timeit'] = True
        func.__dict__['precision'] = int(precision)
        return func
    return inject


class ResultUnit:
    def __init__(self, **kwargs):
        self.case_num = 0
        self.func_name = ''
        self.func_object = None
        self.args = None
        self.kwargs = None
        self.output = None
        self.duration = 0
        self.__dict__.update(kwargs)

    def __str__(self):
        if hasattr(self.func_object, 'timeit'):
            return (f'({self.duration:.{self.func_object.precision}f}s)'
                    f'{self.output}')
        return f'{self.output}'


FnTempl = """
def test_{func}_case_{case_num}(solution_obj, case{case_num}):
    output = solution_obj.{func}(*case{case_num}.args, **case{case_num}.kwargs)
    assert case{case_num}.assert_fn(output)

"""

OutputTempl = """
def test_{func}_case_{case_num}(solution_obj, case{case_num}):
    output = solution_obj.{func}(*case{case_num}.args, **case{case_num}.kwargs)
    assert output == case{case_num}.assert_output

"""


class TestKind(Enum):
    Null = 0
    Output = 1
    WithFn = 2


class Testcase:
    def __init__(self, args, kwargs):
        self.args = list(args)
        self.kwargs = kwargs
        self.assert_output = None
        self.assert_fn = None
        self._test_kind = TestKind.Null

    def __str__(self):
        args = [str(arg) for arg in self.args]
        for k, v in self.kwargs:
            args.append(f"{k}={v}")
        sig = ','.join(args)
        return f"Testcase({shorten(sig, 100)})"

    def __repr__(self):
        return f'<{self}>'

    def test_kind(self):
        return self._test_kind

    def assert_equal(self, x):
        self.assert_output = x
        self._test_kind = TestKind.Output
        return self

    def assert_true_with(self, fn):
        self.assert_fn = fn
        self._test_kind = TestKind.WithFn
        return self


class Solution:
    def __init__(self):
        self.solutions = [item for item in self.__class__.__dict__.values()
                          if hasattr(item, 'solution')]
        self.nontest_cases = []
        self.test_cases = []
        self.context = Context

    def __str__(self):
        n = len(self.solutions)
        plural = 's' if n > 0 else ''
        return f"{self.__class__.__name__} with {n} solution{plural}"

    def __repr__(self):
        return f"<{self}>"

    def set_context(self, context_cls):
        self.context = context_cls

    def case(self, *args, **kwargs):
        args, kwargs = self.context.transform_args(args, kwargs)
        return Testcase(deepcopy(args), deepcopy(kwargs))

    def add_case(self, case):
        if case.test_kind() == TestKind.Null:
            self.nontest_cases.append(case)
        else:
            self.test_cases.append(case)

    # deprecated method. use add_case instead.
    def add_args(self, *args, **kwargs):
        self.add_case(self.case(*args, **kwargs))

    def _run_solution(self, solution, args, kwargs):
        ags, kws = deepcopy(args), deepcopy(kwargs)
        t1 = perf_counter()
        output = solution.__call__(self, *ags, **kws)
        duration = perf_counter() - t1
        return output, duration

    def run_cases_to_table(self):
        result_by_case = []
        for i, case in enumerate(self.nontest_cases):
            case_row = []
            for f in self.solutions:
                output, duration = self._run_solution(
                    f, case.args, case.kwargs)
                r = ResultUnit(
                    case_num=i,
                    func_name=f.__name__,
                    func_object=f,
                    args=case.args,
                    kwargs=case.kwargs,
                    output=output,
                    duration=duration)
                case_row.append(r)
            result_by_case.append(case_row)

        # draw table
        if not result_by_case:
            return
        table_settings = config.get('table')
        table = Table(**table_settings)
        header = ['']
        header.extend([func.__name__ for func in self.solutions])
        table.add_header(header)
        for case_num, case_row in enumerate(result_by_case):
            row = [f'case {case_num}']
            row.extend(case_row)
            table.add_row(row)
        print(table)

    def run_cases_to_test(self):
        if not self.test_cases:
            return
        test_code = StringIO()
        for i, case in enumerate(self.test_cases):
            for func in self.solutions:
                if case.test_kind() == TestKind.WithFn:
                    templ = FnTempl
                elif case.test_kind() == TestKind.Output:
                    templ = OutputTempl
                code = templ.format(case_num=i, func=func.__name__)
                test_code.write(code)

        plugin_text = """
        class FixturePlugin:
            def __init__(self, q_instance):
                self.q = q_instance

            @pytest.fixture(scope='module')
            def solution_obj(self):
                return self.q
        """

        case_text = """
            @pytest.fixture(scope='module')
            def case{case_num}(self):
                return self.q.test_cases[{case_num}]
        """
        for i in range(len(self.test_cases)):
            plugin_text += case_text.format(case_num=i)

        exec(dedent(plugin_text))

        fixture_class = locals()['FixturePlugin']

        project_dir = Path(inspect.getfile(self.__class__)).parent.parent
        test_filename = f'test_{self.__class__.__name__}.py'
        test_file = project_dir / test_filename
        with open(test_file, 'w') as f:
            f.write(test_code.getvalue())
        try:
            pytest.main(['-q', str(test_file)], plugins=[fixture_class(self)])
        finally:
            os.remove(test_file)

    def run(self):
        self.run_cases_to_table()
        self.run_cases_to_test()



import sys
import json
import logging
from types import SimpleNamespace


LOG = logging.getLogger(__name__)
Debug = LOG.debug

try:
    import numpy as np
    import seaborn as sns
    import matplotlib.pyplot as plt
except ImportError as e:
    print("plot a heatmap needs package 'seaborn', try to install it first:")
    print("    $ pip install seaborn")
    Debug(e)
    sys.exit(1)


from leezy.crawler import Net
from leezy.config import Urls


MAX_NUM = 3000


class DataFeeder:
    def __init__(self):
        self.net = Net()
        self.raw = self.net.get(Urls.api_problems_algo()).json()
        self.metadata = SimpleNamespace(**{
            'user_name': self.raw['user_name'],
            'num_solved': self.raw['num_solved'],
            'ac_easy': self.raw['ac_easy'],
            'ac_medium': self.raw['ac_medium'],
            'ac_hard': self.raw['ac_hard'],
        })

        ac = [None] * MAX_NUM
        attempt = [None] * MAX_NUM
        locked = [None] * MAX_NUM
        for stat in self.raw['stat_status_pairs']:
            try:
                front_id = int(stat['stat']['frontend_question_id'])
            except ValueError:
                continue
            if front_id > MAX_NUM:
                pass
            ac[front_id] = stat['status'] == 'ac'
            locked[front_id] = stat['paid_only']
            attempt[front_id] = stat['status'] == 'notac'

        max_id = MAX_NUM - 1
        while ac[max_id] is None:
            max_id -= 1
        max_id += 1
        ac = ac[:max_id]
        attempt = attempt[:max_id]
        locked = locked[:max_id]

        def none2false(arr):
            for i, x in enumerate(arr):
                if x is None:
                    arr[i] = False
            return arr

        self.max_id = max_id
        self.ac = none2false(ac)
        self.attempt = none2false(attempt)
        self.locked = none2false(locked)

class SNSPlotter:
    def __init__(self, feeder):
        self.feeder = feeder
        self.max_id = feeder.max_id
        self.ac_vector = feeder.ac

    def plot(self):
        sns.set()

        row_cnt, remaining = divmod(self.max_id, 100)
        row_cnt += 1
        size = row_cnt * 100
        vector = self.ac_vector + [False] * (size - self.max_id)
        data = np.reshape(vector, (row_cnt, 100))
        mask = np.zeros_like(data)
        mask[0, 0] = True
        mask[row_cnt-1, remaining:] = True

        with sns.axes_style("white"):
            f, ax = plt.subplots(figsize=(16, 9), dpi=100)
            sns.heatmap(
                data, mask=mask, ax=ax,
                center=0.6,
                linewidths=2,
                yticklabels=list(range(0, size, 100)),
                square=True,
                cbar=False,
            )

        plt.show()



from .config import Config
from .errors import ConfigError
import pytest
from pathlib import Path
from unittest import mock



test_config_file = '.leeezy'


@pytest.fixture(scope='function')
def config():
    Config._instance = None
    Config.init.__globals__['CONFIG_FILE'] = test_config_file
    config = Config()
    return config


def test_config_get_default(config):
    assert config.get('core.workdir') == '.'
    assert config.get('core.zone') == 'cn'

    core_dict = config.get('core')
    assert core_dict['workdir'] == '.'
    assert core_dict['zone'] == 'cn'


def test_config_put(config):
    token = 'abcdefg'
    expires = 12432552
    config.put('session.token', token)
    config.put('session.expires', expires)

    sess = config.get('session')
    assert sess['expires'] == expires

    file = Path(test_config_file)
    content = file.read_text()
    assert token in content
    assert str(expires) in content

    Config._instance = None
    c = Config()
    assert config.get('session.token') == token
    file.write_text('{}')


def test_config_patch(config):
    config.patch('universe.mystery.answer', 42)
    assert config.get('universe.mystery.answer') == 42
    assert len(config.file_data) == 0


def test_config_patch_error(config):
    config.patch('a.b.c', 42)
    with pytest.raises(ConfigError):
        config.patch('a.b.c.d', 41)



import re
from enum import Enum
from .templite import Templite


class TemplateType(Enum):
    Normal = 0  # single function solution
    Design = 1  # design a data structure
    Unkown = 2  # rare situation


NormalTempl = """from leezy import solution, Solution
{% if tree_context -%}
from leezy.assists import TreeContext
{% endif -%}
{% if linkedlist_context -%}
from leezy.assists import LinkedListContext
{% endif -%}


class Q{{id_}}(Solution):
    @solution
    {{defs.0}}
        pass


def main():
    q = Q{{id_}}()
{% if tree_context -%}
    q.set_context(TreeContext)
{% endif -%}
{% if linkedlist_context -%}
    q.set_context(LinkedListContext)
{% endif -%}
    q.add_case(q.case({{testcase}}))
    q.run()


if __name__ == '__main__':
    main()
"""

DesignTempl = """
{{ code_snippet }}

def main():
    {{inst}} = {{clss.0}}({{init_args}})
    operations = {{testcase.0}}
    oprands = {{testcase.1}}
    for opt, opd in zip(operations, oprands):
        if hasattr({{inst}}, opt):
            print(getattr({{inst}}, opt).__call__(*opd))


if __name__ == '__main__':
    main()
"""

UnkownTempl = """
# unrecoginzed solution pattern, leave it on your own

{{code_snippet}}

def main():
    pass

if __name__ == '__main__':
    main()
"""


class Render:
    def __init__(self, problem):
        self.problem = problem

    def template_type(self):
        """
        infer template type by observing number and name of `def` `class`
        """
        defs, clss = self.extract_def_and_cls()
        testcase = self.problem.sample_testcase
        if len(defs) == 1:
            return TemplateType.Normal
        elif len(clss) == 1 and clss[0] != 'Solution' and len(testcase) == 2:
            return TemplateType.Design
        else:
            return TemplateType.Unkown

    def extract_def_and_cls(self):
        code_snippet = self.problem.code_snippet
        re_def = re.compile(r'^\s*(def .*:)\n', re.MULTILINE)
        re_class = re.compile(r'^\s*class ([_\w\d]+)\(?.*\)?:\n', re.MULTILINE)
        clss = re_class.findall(code_snippet)
        defs = re_def.findall(code_snippet)
        return defs, clss

    def render(self):
        tmpl_type = self.template_type()
        defs, clss = self.extract_def_and_cls()
        problem = self.problem
        context = {
            'defs': defs,
            'clss': clss,
        }

        code = ''
        if tmpl_type == TemplateType.Normal:
            context.update({
                'tree_context': self.problem.context == 'tree',
                'linkedlist_context': self.problem.context == 'linkedlist',
                'id_': problem.loc_id if problem.loc_id.isdigit() else 'Solution',
                'testcase': ", ".join(repr(x) for x in problem.sample_testcase)
            })
            t = Templite(NormalTempl)
            code = t.render(context)
        elif tmpl_type == TemplateType.Design:
            testcase = problem.sample_testcase
            init_args = ''
            if testcase[0][0] == clss[0]:  # first operation is initializing
                init_args = ', '.join(repr(x) for x in testcase[1][0])
                testcase[0] = testcase[0][1:]
                testcase[1] = testcase[1][1:]
            context.update({
                'code_snippet': problem.code_snippet,
                'inst': clss[0].lower(),
                'init_args': init_args,
                'testcase': [repr(case) for case in testcase],
            })
            t = Templite(DesignTempl)
            code = t.render(context)
        else:
            t = Templite(UnkownTempl)
            code = t.render(dict(code_snippet=problem.code_snippet))

        return code.replace('(object):', ':')



# http://aosabook.org/en/500L/a-template-engine.html

import re
import keyword


class TempliteSyntaxError(ValueError):
    """编译时，语法无法识别"""
    pass


class TempliteRenderError(ValueError):
    """运行时，do_dots出现异常"""
    pass


class CodeBuilder:
    INDENT = 4

    def __init__(self, init_indent=0):
        self.indent_lv = init_indent
        self.code = []

    def indent(self):
        self.indent_lv += self.INDENT

    def dedent(self):
        self.indent_lv -= self.INDENT

    def add_line(self, line):
        self.code.extend([' '*self.indent_lv, line, '\n'])

    def add_empty_line(self):
        self.code.append('\n')

    def add_section(self):
        """添加一个子CodeBuilder

        功能上等效于增加一个占位符，构建完后续代码后才能确定当前section的内容
        例子是绑定局部变量之前，必须解析后续代码才能直到有那些局部变量被使用了
        """
        section = CodeBuilder(self.indent_lv)
        self.code.append(section)
        return section

    def __str__(self):
        return ''.join(str(c) for c in self.code)

    def globals_after_exec(self):
        """执行代码，获取global名字空间

        Raises:
            SyntaxError: 如果执行时indent_lv不为0或者代码本身存在语法错误
        """
        if self.indent_lv != 0:
            raise SyntaxError("unbalanced CodeBuilder")
        source_code = str(self)
        global_namespace = {}
        exec(source_code, global_namespace)
        return global_namespace


class Templite:
    def __init__(self, text, *context):
        self.text = text
        self.context = {}
        for cxt in context:
            self.context.update(cxt)

        # context_variables = all_variables - native_variables
        self.all_variables = set()
        self.native_variables = set()

        code = self._build(self._tokens(self.text))
        self._co_code = code
        self._render_func = code.globals_after_exec()['render_template']

    def render(self, context=None):
        context = context or {}
        context.update(self.context)
        return self._render_func(context, self._do_dots)

    def _tokens(self, text):
        # 单独切分token，之后如果要处理转义在这里执行？
        tokens = re.split(r"(?s)({{.*?}}|{%.*?%}|{#.*?#})", text)
        return tokens

    def _build(self, tokens):
        code = CodeBuilder()
        code.add_line('def render_template(context, do_dots):')
        code.indent()
        code.add_line('result = []')
        # 速度优化之一，查用名字固定到local空间，减少查找
        code.add_line('append_result = result.append')
        code.add_line('expand_result = result.extend')
        code.add_line('to_str = str')
        # context的变量初始化，明确哪些变量被使用后再添加赋值语句
        context_vars_section = code.add_section()

        buffer = []

        # 速度优化之二，对于非控制语句，填充入buffer，一次性由extend拓展，减少扩容操作
        def flush():
            if len(buffer) == 0:
                return
            elif len(buffer) == 1:
                code.add_line(f'append_result({buffer[0]})')
            else:
                code.add_line(f'expand_result([{", ".join(buffer)}])')
            buffer.clear()

        squash = False
        ops_stack = []
        for token in tokens:
            if token.startswith('{'):
                # '{' or '%' or '#'
                tag = token[1]
                content = token[2:-2].strip()
                # 每次遇到控制语句，都更新squash，指示下一个文本值如何处理
                # 注意token的顺序一定是 文本 - 控制 - 文本 - 控制 - ... 的顺序
                if content[-1] == '-':
                    squash = True
                    content = content[:-1].strip()
                else:
                    squash = False
                if tag == '#':
                    continue
                elif tag == '{':
                    buffer.append(f'to_str({self._expr(content)})')
                elif tag == '%':
                    flush()
                    if content.startswith('if'):
                        parts = content.split()
                        if len(parts) != 2:
                            self._syntax_error("Don't understand if", content)
                        var_code = self._expr(parts[1])
                        code.add_line(f'if {var_code}:')
                        code.indent()
                        ops_stack.append('if')
                    elif content.startswith('for'):
                        parts = content.split()
                        if len(parts) != 4 or parts[2] != 'in':
                            self._syntax_error("Don't understand for", content)
                        self._record_variable(parts[1], self.native_variables)
                        iter_var = self._expr(parts[3])
                        code.add_line(f'for c_{parts[1]} in {iter_var}:')
                        code.indent()
                        ops_stack.append('for')
                    elif content.startswith('end'):
                        parts = content.split()
                        if len(parts) != 1:
                            self._syntax_error("Don't understand end", content)
                        act = content[3:]
                        if len(ops_stack) == 0:
                            self._syntax_error("Too many ends", token)
                        if ops_stack.pop() != act:
                            self._syntax_error("unmatched end tag", token)
                        code.dedent()
                    else:
                        self._syntax_error('Unkown action', content)
                else:
                    self._syntax_error("Unkown tag", token)

            elif token:
                if squash:
                    token = self._delete_first_leading_newline(token)
                # 字面值，直接加入buffer中
                if token:
                    buffer.append(repr(token))

        if ops_stack:
            self._syntax_error("No matched end tag", ops_stack[-1])

        for cxt_var in self.all_variables - self.native_variables:
            context_vars_section.add_line(f"c_{cxt_var} = context[{cxt_var!r}]")

        flush()
        code.add_line('return "".join(result)')
        code.dedent()
        return code

    def _delete_first_leading_newline(self, token):
        """删除先导空白字符中的第一个换行符

        Examples::
        >>> _delete_first_leading_newline('  \n    \ntoken content\n')
        '    \ntoken content\n'
        >>> _delete_first_leading_newline('  token content\n')
        '  token content\n'
        """
        idx = token.find('\n')
        if token[:idx+1].isspace():
            token = token[idx+1:]
        return token

    def _syntax_error(self, msg, thing):
        raise TempliteSyntaxError(f"{msg}: {thing}")

    def _expr(self, expr):
        """展开表达式

        Examples:
        >>> _expr('obj1.attr.val|obj2.dump')
        do_dots(c_obj2, 'dump')(do_dots(c_obj1, 'attr', 'val'))

        Args:
            expr: 可能混合 . 和 | 符号的字符串, 出现在 {{ expr }} {% if expr %}
                    {% for x in expr %}语句中

        Returns:
            code: 展开后的可执行的源码

        Raises:
            TempliteSyntaxError: 如果展开过程中发现不合法的变量名
        """
        # TODO: 支持|左右存在空格？
        if '|' in expr:
            parts = expr.split('|')
            code = self._expr(parts[0])
            for p in parts[1:]:
                code = "{}({})".format(self._expr(p), code)
        elif '.' in expr:
            # 此时expr中不存在'|'
            parts = expr.split('.')
            code = self._expr(parts[0])
            tag_args = ', '.join(repr(p) for p in parts[1:])
            code = f"do_dots({code}, {tag_args})"
        else:
            # 单独一个变量名，如果不是，交给record抛出异常
            self._record_variable(expr, self.all_variables)
            code = f"c_{expr}"
        return code

    def _record_variable(self, name, var_set):
        if not name.isidentifier() or keyword.iskeyword(name):
            self._syntax_error("Not a valid name", name)
        var_set.add(name)

    def _do_dots(self, obj, *tags):
        """运行时解析dot语法

        模板中的dot语法共三种行为::
            1. 方法调用， obj.method()
            2. 取attribute， obj.attr
            3. index， obj[tag]，比如列表和字典的索引行为
        前两种行为优先级更高
        """
        for tag in tags:
            try:
                obj = getattr(obj, tag)
            except AttributeError:
                try:
                    obj = obj[tag]
                except (TypeError, KeyError):
                    if tag.isdigit():
                        try:
                            obj = obj[int(tag)]
                        except (TypeError, KeyError, IndexError):
                            raise TempliteRenderError(
                                f"Can't resolve dot: {obj}.{tag}")
                    else:
                        raise TempliteRenderError(
                            f"Can't resolve dot: {obj}.{tag}")
            else:
                if callable(obj):
                    obj = obj()
        return obj



import json
import logging
from pathlib import Path
from collections import abc
from functools import partial
from datetime import datetime
from types import SimpleNamespace
from requests.cookies import RequestsCookieJar

from leezy.errors import ConfigError


__all__ = ['config']


DEFAULT = {
    "table": {
        "max_col_width": 30,
        "max_content_length": 100
    },
    "core": {
        "workdir": ".",
        "zone": "cn"
    },
    "log": {
        "level": "WARNING"
    },
    "timeout": {
        "submit": 10,
        "net": 5
    }
}


CHECK_FUNCTIONS = {
    "table.max_col_width": int,
    "table.max_content_length": int
}

CONFIG_FILE = '~/.leezy'


class Config:
    """
    >>> config = Config()
    >>> config.put('answer', 42)
    >>> config.get('answer')
    42
    >>> config.put('user.name', 'x')
    >>> config.put('user.email', 'x@example.com')
    >>> config.get('user')
    {'name': 'x', 'email': 'x@example.com'}
    >>> config.reset()
    """
    _instance = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)
            cls.init(cls._instance)
        return cls._instance

    def init(self):
        self.cfg_path = Path(CONFIG_FILE).expanduser()
        self.default_data = DEFAULT
        self.mem_data = {}
        try:
            content = self.cfg_path.read_text(encoding='utf8')
            self.file_data = json.loads(content)
        except FileNotFoundError:
            self.reset()

    def reset(self):
        self.mem_data = {}
        self.file_data = {}
        self.commit()

    def commit(self):
        with open(self.cfg_path, 'w') as f:
            json.dump(self.file_data, f, indent=2)

    def _get(self, src_data, key):
        next_item = src_data
        for part in key.split('.'):
            try:
                next_item = next_item[part]
            except KeyError:
                raise
        return next_item

    def _get_from_chain(self, key, src_chain):
        for src_data in src_chain:
            try:
                ret = self._get(src_data, key)
            except KeyError:
                pass
            else:
                return ret
        else:
            raise ConfigError(f"config: {key!r} is not found")

    def get(self, key):
        return self._get_from_chain(key, (self.mem_data,
                                          self.file_data,
                                          self.default_data))

    def _get_all(self, mapping, prefix):
        for key, value in mapping.items():
            next_prefix = prefix+'.'+key if prefix else key
            if isinstance(value, abc.Mapping):
                yield from self._get_all(value, next_prefix)
            else:
                yield next_prefix, value

    def get_all_file_data(self, prefix=''):
        yield from self._get_all(self.file_data, prefix)

    def _put(self, key, value, src_data):
        check_fn = CHECK_FUNCTIONS.get(key, None)
        if check_fn:
            try:
                check_fn(value)
            except:
                raise ConfigError(f"config: {value!r} is invalid for {key!r}")
        parts = key.split('.')
        next_item = src_data
        for part in parts[:-1]:
            if part not in next_item:
                next_item[part] = {}
            next_item = next_item[part]

        if not isinstance(next_item, dict):
            path = '.'.join(parts[:-1])
            raise ConfigError(f"config: failed to assign {key!r}, "
                              f"because {path!r} is not a map")
        next_item[parts[-1]] = value

    def put(self, key, value):
        """update config entry and persist the data"""
        self._put(key, value, self.file_data)
        self.commit()

    def patch(self, key, value):
        """update config entry in memory"""
        self._put(key, value, self.mem_data)

    def _del(self, key, src_data):
        parts = key.split('.')
        next_item = src_data
        for part in parts[:-1]:
            try:
                next_item = next_item[part]
            except KeyError:
                raise
        try:
            del next_item[parts[-1]]
        except KeyError:
            raise

    def delete(self, key):
        deleted = 0
        for src_data in (self.mem_data, self.file_data):
            try:
                self._del(key, src_data)
                deleted += 1
            except KeyError:
                pass
        if deleted == 0:
            raise ConfigError(f"config: {key!r} is not found")
        self.commit()
        return deleted


class SessionToken:
    def init(self):
        """use global `config` to initial itself like Urls.init

        call this method when config is all set.
        """
        # let __main__ to initial session_token
        zone = self.config.get('core.zone')
        if zone == 'cn':
            self.token_path = 'session.cn.token'
            self.expires_path = 'session.cn.expires'
            self.csrf_path = 'session.cn.csrf'
            self.domain = '.leetcode-cn.com'
        elif zone == 'us':
            self.token_path = 'session.us.token'
            self.expires_path = 'session.us.expires'
            self.csrf_path = 'session.us.csrf'
            self.domain = 'leetcode.com'
        else:
            raise ConfigError(f'Unrecognized zone {zone!r}')
        try:
            self.token = config.get(self.token_path)
            self.expires = config.get(self.expires_path)
        except ConfigError:
            pass
        try:
            self.csrf = config.get(self.csrf_path)
        except ConfigError:
            pass

    def __init__(self, config):
        self.token = None
        self.expires = None
        self.csrf = None
        self.domain = None
        # test might use this field
        self.config = config

    def is_existed(self):
        return self.expires is not None and self.token is not None

    def is_expired(self):
        return datetime.timestamp(datetime.now()) > self.expires

    def get_token(self):
        jar = RequestsCookieJar()
        if self.token:
            jar.set('LEETCODE_SESSION', self.token, domain=self.domain)
        return jar

    def get_csrf(self):
        jar = RequestsCookieJar()
        if self.csrf:
            jar.set('csrftoken', self.csrf, domain=self.domain)
        return jar

    def store_token(self, token, expires):
        self.token = token
        self.expires = expires
        self.config.put(self.token_path, token)
        self.config.put(self.expires_path, expires)

    def try_update_csrf(self, r):
        if 'csrftoken' in r.cookies:
            self.store_csrf(r.cookies['csrftoken'])
            return True
        return False

    def store_csrf(self, csrf):
        # actually, csrftoken's lifetime is very long
        # it may be more efficient to lower the frequency of updating
        if csrf != self.csrf:
            self.csrf = csrf
            self.config.put(self.csrf_path, csrf)


class Urls:
    PORTAL = None

    @classmethod
    def init(cls, config):
        """use global `config` to get zone info to initial itself

        call this method when config is all set.
        """
        zone = config.get('core.zone')
        if zone == 'cn':
            cls.PORTAL = 'https://leetcode-cn.com'
        elif zone == 'us':
            cls.PORTAL = 'https://leetcode.com'
        else:
            raise ConfigError(f'Unrecognized zone {zone!r}')

    @staticmethod
    def portal():
        return Urls.PORTAL

    @staticmethod
    def graphql():
        return f"{Urls.PORTAL}/graphql/"

    @staticmethod
    def api_problems_algo():
        return f"{Urls.PORTAL}/api/problems/algorithms/"

    @staticmethod
    def api_problems_all():
        return f"{Urls.PORTAL}/api/problems/all/"

    @staticmethod
    def problem_home(slug_title):
        return f"{Urls.PORTAL}/problems/{slug_title}/"

    @staticmethod
    def problem_discussion(slug_title):
        # you may don't want to see discussion in cn for now
        return f"https://leetcode.com/problems/{slug_title}/discuss/"

    @staticmethod
    def problem_submit(slug_title):
        # POST  https://leetcode-cn.com/problems/two-sum/submit/
        return f"{Urls.PORTAL}/problems/{slug_title}/submit/"

    @staticmethod
    def submission_detail(sub_id):
        # GET https://leetcode-cn.com/submissions/detail/53947058/
        return f"{Urls.PORTAL}/submissions/detail/{sub_id}"

    @staticmethod
    def submission_check(sub_id):
        # GET https://leetcode-cn.com/submissions/detail/53947058/check/
        return f"{Urls.PORTAL}/submissions/detail/{sub_id}/check/"


config = Config()
session_token = SessionToken(config)



from unittest import mock
import pytest
import textwrap
import re

from .templite import (Templite, CodeBuilder,
                      TempliteSyntaxError, TempliteRenderError)


@pytest.fixture
def empty_templite():
    return Templite("")


def test_codebuilder():
    builder = CodeBuilder()
    builder.add_line('def hi():')
    builder.indent()
    builder.add_line('print("hello, world")')
    builder.dedent()
    builder.add_empty_line()
    builder.add_line('DEBUG = True')
    global_ns = builder.globals_after_exec()
    assert 'hi' in global_ns and 'DEBUG' in global_ns
    assert global_ns['DEBUG'] is True


def test_codebuilder_unbalanced():
    builder = CodeBuilder()
    builder.add_line('def hi():')
    builder.indent()
    builder.add_line('print("hello, world")')
    with pytest.raises(SyntaxError):
        builder.globals_after_exec()


def test_expr(empty_templite):
    t = empty_templite
    code = t._expr("obj1.attr.val|obj2.dump")
    assert code == "do_dots(c_obj2, 'dump')(do_dots(c_obj1, 'attr', 'val'))"
    assert len(t.all_variables) == 2
    assert 'obj1' in t.all_variables and 'obj2' in t.all_variables


@pytest.mark.parametrize('expr', ['51day', 'my-goods',
                                  'flag and flag1', 'class.upper'])
def test_invalid_expr(empty_templite, expr):
    t = empty_templite
    with pytest.raises(TempliteSyntaxError):
        t._expr(expr)
    assert len(t.all_variables) == 0


def test_do_dots(empty_templite):
    t = empty_templite
    assert 'N' == t._do_dots('name', 'upper', '0')
    import sys
    assert 'Win32' == t._do_dots(sys, 'platform', 'title')


@pytest.mark.parametrize(
    'args', [['day', 'nomethod'], [mock, 'nomock()'], [[1, 2], '3']])
def test_do_dots_caught_fail(empty_templite, args):
    t = empty_templite
    with pytest.raises(TempliteRenderError):
        t._do_dots(args[0], *args[1:])


@pytest.mark.parametrize(
    'text',
    [
        '{& unkown tag &}',  # 未知tag
        '{% if a %} skr ',  # 没有end
        'skr {% endif %}',  # 单独end
        '{% if a %} skr {% endfor %}',  # 错误匹配
        # 错误语法
        '{% if a: %} skr {% endif %}',
        '{% if a %} skr {% end if %}',
        '{% for a in p | transfer %} {{a}} {% endfor %}',  # 多余的空格
        # 未支持语法
        '{% while a %} skr {% endwhile %}'
        '{{ name().upper }}',
        '{% if a and b %} skr {% endif %}',
        '{% for a, b in a.enumerate %} skr {% endfor %}',
    ]
)
def test_syntax_error_actions(text):
    with pytest.raises(TempliteSyntaxError):
        Templite(text)


TEXT_VAR = """
<p>Welcome, {{ user_name.upper|json_dump }}!</p>
<p>Ula, {{ user_age }}!</p>
{# this is a comment #}
"""

TEXT_IF = """
-----------Cyberpunk-------------
{% if position %}
Mid-night Bar{% if time %} · 2077 {% endif %}
{% endif %}
...
"""

TEXT_FOR = """
Points:
{% for x in xs -%}
{% for y in ys -%}
    ({{ x }}, {{ y }})
{% endfor -%}
{% endfor -%}
"""


def test_tokens(empty_templite):
    t = empty_templite
    assert len(t._tokens(TEXT_VAR)) == 7
    assert len(t._tokens(TEXT_IF)) == 9
    assert len(t._tokens(TEXT_FOR)) == 13


def assert_executable_codebuilder(code):
    global_ns = code.globals_after_exec()
    assert 'render_template' in global_ns
    assert callable(global_ns['render_template'])


def test_build_var(empty_templite):
    t = empty_templite
    tokens = t._tokens(TEXT_VAR)
    code = t._build(tokens)
    code_str = str(code)
    # 纯变量，使用buffer写入，因此有expand_result的调用
    assert 'expand_result([' in code_str
    assert "to_str(c_json_dump(do_dots(c_user_name, 'upper')))" in code_str
    # 可以成功执行，生成代码没有语法错误
    assert_executable_codebuilder(code)


def test_build_if(empty_templite):
    t = empty_templite
    tokens = t._tokens(TEXT_IF)
    code = t._build(tokens)
    code_str = str(code)
    assert 'if c_position:' in code_str
    assert 'if c_time' in code_str
    assert_executable_codebuilder(code)


def test_build_for(empty_templite):
    t = empty_templite
    tokens = t._tokens(TEXT_FOR)
    code = t._build(tokens)
    code_str = str(code)
    assert 'for c_x in c_xs:' in code_str
    assert 'for c_y in c_ys:' in code_str
    assert_executable_codebuilder(code)


def test_render_var():
    import json
    from functools import partial
    t = Templite(TEXT_VAR, {
        'format_price': lambda x: f"{x:.2f}",
        'json_dump': partial(json.dumps, ensure_ascii=False)
    })
    result = t.render({
        'user_name': 'Gimo',
        'user_age': 42,
    })
    expected = """
    <p>Welcome, "GIMO"!</p>
    <p>Ula, 42!</p>

    """
    assert textwrap.dedent(expected) == result

    result = t.render({
        'user_name': '梭哈',
        'user_age': 33,
    })
    expected = """
    <p>Welcome, "梭哈"!</p>
    <p>Ula, 33!</p>

    """
    assert textwrap.dedent(expected) == result


def test_render_if():
    t = Templite(TEXT_IF)
    result = t.render({
        'position': True,
        'time': True,
    })
    assert "Mid-night Bar · 2077" in result

    result = t.render({
        'position': True,
        'time': False,
    })
    assert "Mid-night Bar" in result
    assert "2077" not in result

    result1 = t.render({
        'position': False,
        'time': False,
    })
    result2 = t.render({
        'position': False,
        'time': True,
    })
    assert result1 == result2
    assert "Bar" not in result1 and "2077" not in result1


def test_render_for():
    t = Templite(TEXT_FOR)
    point_re = re.compile(r"\(.+?, .+?\)", re.MULTILINE)
    result = t.render({
        'xs': [1, 2],
        'ys': [3, 4],
    })
    assert len(point_re.findall(result)) == 4

    result = t.render({
        'xs': [1, 2],
        'ys': [3, 4, 5],
    })
    assert len(point_re.findall(result)) == 6


def test_render_squash():
    t = Templite(TEXT_FOR)
    result = t.render({
        'xs': [1, 2],
        'ys': [3, 4],
    })
    assert len(result.split('\n')) == 7

    result = t.render({
        'xs': [1, 2],
        'ys': [3, 4, 5],
    })
    assert len(result.split('\n')) == 9




import sys
import logging
import requests

LOG = logging.getLogger(__name__)
Info = LOG.info
Debug = LOG.debug
Warn = LOG.warning


def show_error_and_exit(err):
    print(err)
    sys.exit(2)


def raise_for_status(response, description):
    try:
        response.raise_for_status()
    except requests.exceptions.HTTPError as e:
        Debug(response.text)
        Debug("\nrequst headers:\n" + str(response.request.headers))
        if e.response.status_code == 403:
            # avoid loop forever during import-time
            from leezy.config import config
            zone = config.get('core.zone')
            assert zone == 'cn' or zone == 'us'
            config.delete("session."+zone)
            raise FetchError(
                "Account authentication failed, you might try again and sign in")
        raise FetchError(description, e)


class LeezyError(Exception):
    def __init__(self, description="", cause=None):
        self.description = description
        self.cause = cause

    def __str__(self):
        display = f'A `{self.__class__.__name__}` occurred'
        if self.description:
            display += '\n information: ' + self.description + '\n'
        if self.cause:
            display += 'caused by:\n'
            display += str(self.cause) + '\n'
        return display


class LoginError(LeezyError):
    pass


class NetworkError(LeezyError):
    pass


class FetchError(LeezyError):
    pass


class NotFound(LeezyError):
    pass


class Locked(LeezyError):
    pass


class ConfigError(LeezyError):
    pass



import re
from setuptools import setup

with open('leezy/__main__.py', encoding='utf8') as f:
    VERSION = re.findall(
        r'VERSION\s*=\s*[\'\"]([\.\d]+)[\'\"]\s*$',
        f.read(), re.MULTILINE)[0]

with open("README.md", 'r', encoding='utf8') as f:
    long_description = f.read()

setup(
    name='leezy',
    version=VERSION,
    description='leezy: leetcode helper for the lazy',
    license="MIT",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author='aptend',
    author_email='aptend@hotmail.com',
    url='https://github.com/aptend/leezy',
    entry_points={
        'console_scripts': [
            'leezy = leezy.__main__:dummy_main'
        ]
    },
    classifiers=[
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Topic :: Text Processing',
        'Topic :: Utilities'
    ],
    packages=['leezy'],
    install_requires=['requests>=2.18.0', 'pytest>=5.1.3'],
    python_requires='>=3.6'
)



import sys
import logging
import argparse
import subprocess

from leezy.crawler import Problem
from leezy.config import config, session_token, Urls

from leezy.errors import show_error_and_exit, LeezyError

import traceback

LOG = logging.getLogger(__name__)
Info = LOG.info
Debug = LOG.debug
Warn = LOG.warning

VERSION = '0.5.3'


def show_uncaught_exc(e):
    print(f'Uncaught Exception: {e!r}')
    Debug(traceback.format_exc(limit=10))


def _exit(msg):
    print('leezy: ' + msg)
    sys.exit(1)


def expand_ids(ids_arg):
    if len(ids_arg) == 1 and ids_arg[0].count('-') == 1:
        # handle input like '40-43'
        s, e = ids_arg[0].split('-')[:2]
        try:
            s, e = int(s), int(e)
        except ValueError:
            _exit('a-b is valid only when a and b are both numbers')
        return [str(x) for x in range(s, e+1)]
    else:
        return ids_arg


def parse_solution_pos(s):
    # handle input like 2@42
    parts = s.strip().split('@')
    N = len(parts)
    if N == 0 or N > 2:
        _exit(f'unrecognized soution input: {s!r}')
    if N == 1:
        return (1, s)
    else:
        try:
            sol_num = int(parts[0])
        except ValueError:
            _exit(f'{parts[0]!r} is not a number')
        return (sol_num, parts[1])


parser = argparse.ArgumentParser(
    prog='leezy',
    description='Manage your Python solutions better',
    usage='leezy [options] COMMAND')

parser.add_argument('-V', '--version',
                    action='version',
                    version=VERSION)
parser.add_argument('--zone',
                    choices=['cn', 'us'],
                    metavar='ZONE',
                    help="'cn' or 'us', default is 'cn'")
parser.add_argument('--dir',
                    help="assign a temporary workdir for this session")
parser.add_argument('-v',
                    action='count',
                    help="verbose, use multiple -vv... to show more log")


subs = parser.add_subparsers(
    title="COMMANDS",
    description="use 'leezy <COMMAND> -h' to see more",
    metavar='-⭐-')


def show(args):
    for pid in expand_ids(args.ids):
        try:
            print(Problem(pid).digest())
        except LeezyError as e:
            show_error_and_exit(e)
        except Exception as e:
            show_uncaught_exc(e)


show_parser = subs.add_parser(
    'show',
    usage=argparse.SUPPRESS,
    help='show basic info of problems',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=r"""examples:
    leezy show 1          show the first problem
    leezy show 1 2 3      show (1st, 2nd, 3rd) problems together
    leezy show 1-3        show (1st, 2nd, 3rd) problems together""")
show_parser.add_argument('ids', nargs='+', help="question ids")
show_parser.set_defaults(func=show)


def pull(args):
    for pid in expand_ids(args.ids):
        try:
            Problem(pid, args.context).pull()
        except LeezyError as e:
            show_error_and_exit(e)
        except Exception as e:
            show_uncaught_exc(e)


pull_parser = subs.add_parser(
    'pull',
    usage=argparse.SUPPRESS,
    help='pull problems to local files',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=r"""examples:
    leezy pull 1                 pull the first problem
    leezy pull 1 2 3             pull (1st, 2nd, 3rd) problems together
    leezy pull 1-3               pull (1st, 2nd, 3rd) problems together
    leezy pull 700 -c tree       pull no.700 and set tree context
    leezy pull 2 -c linkedlist   pull no.2 and set linkedlist context""")

pull_parser.add_argument('ids', nargs='+', help="problem ids")
pull_parser.add_argument('-c', '--context',
                         metavar='',
                         choices=['tree', 'linkedlist'],
                         help="set a context for this problem [tree or linkedlist]")
pull_parser.set_defaults(func=pull)


def run(args):
    try:
        py_path = Problem(args.id).py_path
    except LeezyError as e:
        show_error_and_exit(e)
    except Exception as e:
        show_uncaught_exc(e)

    if not py_path.is_file():
        print(f'File not found: {py_path}')
    else:
        try:
            subprocess.run(['python', str(py_path)],
                           timeout=5,
                           cwd=py_path.parent)
        except FileNotFoundError:
            print('python can\'t be launched by command \'python\'')
        except subprocess.TimeoutExpired:
            print('Timeout(10s). Is there an infinite loop in the solution?')
        except Exception:
            raise


run_parser = subs.add_parser(
    'run',
    usage=argparse.SUPPRESS,
    help='run your solutions, see outputs or test them',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=r"""examples:
    leezy run 1       run the first problem""")
run_parser.add_argument('id', help="problem id")
run_parser.set_defaults(func=run)


def submit(args):
    sol_id, front_id = parse_solution_pos(args.solution)
    try:
        Problem(front_id).submit(sol_id)
    except LeezyError as e:
        show_error_and_exit(e)
    except Exception as e:
        show_uncaught_exc(e)


submit_parser = subs.add_parser(
    'submit',
    usage=argparse.SUPPRESS,
    help='submit your solution to leetcode',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=r"""examples:
    leezy submit 1@1      submit the 1st solution of problem 1
    leezy submit 2@1      submit the 2nd solution of problem 1
    leezy submit 1        same with 1@1, just a shortcut""")

submit_parser.add_argument('solution', help="postion of your solution")
submit_parser.set_defaults(func=submit)


def plot(args):
    from leezy.plot import SNSPlotter, DataFeeder
    SNSPlotter(DataFeeder()).plot()


plot_parser = subs.add_parser(
    'plot',
    usage=argparse.SUPPRESS,
    help='show a heatmap of your all accepted solutions',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description="""examples:
    leezy plot
    """)

plot_parser.set_defaults(func=plot)


def handle_config(args):
    if args.list:
        print('\n'.join('='.join((k, str(v)))
                        for k, v in config.get_all_file_data()))
    elif args.add:
        config.put(args.add[0], args.add[1])
    elif args.delete:
        config.delete(args.delete[0])


config_parser = subs.add_parser(
    'config',
    usage=argparse.SUPPRESS,
    help='manage global configs',
    formatter_class=argparse.RawDescriptionHelpFormatter,
    description=r"""examples:
    leezy config -l                               show all config entries
    leezy config -a core.workdir  D:\leetcode     set workdir
    leezy config -a core.zone cn                  run leezy on cn every time
    leezy config -a log.level debug               run leezy at debug log level
    leezy config -d core                          remove all core settings""")

group = config_parser.add_mutually_exclusive_group()
group.add_argument('-l', '--list',
                   action='store_true',
                   help='show all config entries')
group.add_argument('-d', '--del',
                   nargs=1,
                   metavar='',
                   dest='delete',
                   help='delete settings')
group.add_argument('-a', '--add',
                   nargs=2,
                   metavar='',
                   help='add a <Key Value> setting pair')
config_parser.set_defaults(func=handle_config)

args = parser.parse_args()
if not hasattr(args, 'func'):
    parser.print_help()
else:
    if args.zone is not None:
        config.patch('core.zone', args.zone)
    if args.v is not None:
        if args.v == 1:
            config.patch('log.level', 'info')
        else:
            config.patch('log.level', 'debug')
    if args.dir is not None:
        config.patch('core.workdir', args.dir)

    session_token.init()
    Urls.init(config)

    log_lv = getattr(logging, config.get('log.level').upper())

    def reject_modules(modules, lv=logging.INFO):
        def _reject_filter(record):
            mod_name = record.name
            mod_lv = record.levelno
            for mod in modules:
                if mod_name.startswith(mod) and mod_lv <= lv:
                    return False
            return True
        return _reject_filter

    logging.basicConfig(level=log_lv)
    root = logging.getLogger()
    rej_mat = reject_modules(['matplotlib'])
    for hld in root.handlers:
        hld.addFilter(rej_mat)

    args.func(args)


# this is for setup:entry_points:console_scripts
def dummy_main(*_args, **_kwargs):
    return
